#! -*- coding: utf-8 -*-
from collections import Iterator

from .math_ import euclidean


__all__ = 'CMap', 'PQueue',


class CMap:

    def __init__(self):
        self.vboxes = []

    @property
    def palette(self):
        return [d['color'] for d in self.vboxes]

    def append(self, item):
        self.vboxes.append({'vbox': item, 'color': item.average})

    def __len__(self):
        return len(self.vboxes)

    def nearest(self, color):
        if not self.vboxes:
            raise ValueError('Empty VBoxes!')

        min_d = float('Inf')
        p_color = None
        for vbox in self.vboxes:
            vbox_color = vbox.color
            distance = euclidean(color, vbox_color)
            if min_d > distance:
                min_d = distance
                p_color = vbox.color

        return p_color

    def map(self, color):
        for vbox in self.vboxes:
            if vbox.contains(color):
                return vbox.color

        return self.nearest(color)


class PQueue(Iterator):

    def __init__(self, sorted_key):
        self.sorted_key = sorted_key
        self.items = []

    def __next__(self):
        if not self.items:
            raise StopIteration()

        return self.pop()

    def append(self, item):
        self.items.append(item)
        self.items = sorted(self.items, key=self.sorted_key, reverse=True)

    def pop(self):
        return self.items.pop(0)

    def __len__(self):
        return len(self.items)



from .color import get_color_index
from .constant import SIGBITS, RSHIFT, MAX_ITERATION, FRACT_BY_POPULATIONS
from .region import Vbox
from .struct import CMap, PQueue


def get_histo(colors):
    histo_size = int(1 << (3 * SIGBITS))
    histo = [0 for x in range(histo_size)]
    for color in colors:
        r = color[0] >> RSHIFT
        g = color[1] >> RSHIFT
        b = color[2] >> RSHIFT
        i = get_color_index(r, g, b)
        histo[i] = histo[i] + 1
    return histo


def vbox_from_colors(colors, histo):
    r_colors = []
    g_colors = []
    b_colors = []
    for color in colors:
        r = color[0] >> RSHIFT
        g = color[1] >> RSHIFT
        b = color[2] >> RSHIFT
        r_colors.append(r)
        g_colors.append(g)
        b_colors.append(b)

    return Vbox(min(r_colors),
                max(r_colors),
                min(g_colors),
                max(g_colors),
                min(b_colors),
                max(b_colors),
                histo)


def median_cut(histo, vbox):
    if not vbox.count:
        return None

    rw = vbox.r2 - vbox.r1 + 1
    gw = vbox.g2 - vbox.g1 + 1
    bw = vbox.b2 - vbox.b1 + 1
    maxw = max([rw, gw, bw])
    if vbox.count == 1:
        return vbox.copy,

    tot = 0
    sum_ = 0
    partialsum = {}
    lookaheadsum = {}
    do_cut_color = None
    if maxw == rw:
        do_cut_color = 'r'
        for i in range(vbox.r1, vbox.r2 + 1):
            for j in range(vbox.g1, vbox.g2 + 1):
                for k in range(vbox.b1, vbox.b2 + 1):
                    index = get_color_index(i, j, k)
                    sum_ += histo[index]

            tot += sum_
            partialsum[i] = tot
    elif maxw == gw:
        do_cut_color = 'g'
        for i in range(vbox.g1, vbox.g2 + 1):
            for j in range(vbox.r1, vbox.r2 + 1):
                for k in range(vbox.b1, vbox.b2 + 1):
                    index = get_color_index(j, i, k)
                    sum_ += histo[index]

            tot += sum_
            partialsum[i] = tot
    elif maxw == bw:
        do_cut_color = 'b'
        for i in range(vbox.b1, vbox.b2 + 1):
            for j in range(vbox.r1, vbox.r2 + 1):
                for k in range(vbox.g1, vbox.g2 + 1):
                    index = get_color_index(j, k, i)
                    sum_ += histo[index]

            tot += sum_
            partialsum[i] = tot

    for k, v in partialsum.items():
        lookaheadsum[k] = tot - v

    dim1 = do_cut_color + '1'
    dim2 = do_cut_color + '2'
    dim1_val = getattr(vbox, dim1)
    dim2_val = getattr(vbox, dim2)
    for i in range(dim1_val, dim2_val + 1):
        if partialsum[i] > (tot / 2):
            vbox1 = vbox.copy
            vbox2 = vbox.copy
            left = i - dim1_val
            right = dim2_val - i
            if left <= right:
                d2 = min([dim2_val - 1, ~~(i + int(right / 2))])
            else:
                d2 = max([dim1_val, ~~(i - 1 - int(left / 2))])

            while not partialsum[d2]:
                d2 += 1

            count2 = lookaheadsum[d2]
            while not count2 and (d2 - 1) in partialsum:
                d2 -= 1
                count2 = lookaheadsum[d2]

            setattr(vbox1, dim2, d2)
            setattr(vbox2, dim1, getattr(vbox1, dim2) + 1)
            return vbox1, vbox2


def mmcq(colors, max_color):
    if not isinstance(colors, list) or not colors:
        raise ValueError('`colors` MUST be list '
                         'that contains items not {}'.format(colors))

    if max_color < 2 or max_color > 256:
        raise ValueError('`max_color` MUST be a integer value between '
                         '2 and 256. not {}'.format(max_color))

    pq = PQueue(lambda x: x.count)
    histo = get_histo(colors)
    vbox = vbox_from_colors(colors, histo)
    pq.append(vbox)

    def iter_(lh, target):
        n_color = 1
        n_iter = 0
        while n_iter < MAX_ITERATION:
            vbox = lh.pop()
            if not vbox.count:
                lh.append(vbox)
                n_iter += 1
                continue
            vboxes = median_cut(histo, vbox)
            if not vboxes:
                return None
            lh.append(vboxes[0])
            if len(vboxes) == 2:
                lh.append(vboxes[1])
                n_color += 1
            if n_color >= target:
                return None
            if n_iter > MAX_ITERATION:
                return None
            n_iter += 1

    iter_(pq, FRACT_BY_POPULATIONS * max_color)
    pq2 = PQueue(lambda x: x.volume * x.count)
    for vbox in pq:
        pq2.append(vbox)

    iter_(pq2, max_color - len(pq2))
    cmap = CMap()
    for vbox in pq2:
        cmap.append(vbox)

    return cmap



#! -*- coding: utf-8 -*-
from copy import deepcopy

from .color import get_color_index
from .constant import RSHIFT, SIGBITS


__all__ = 'Vbox',


class Vbox(object):

    def __init__(self, r1, r2, g1, g2, b1, b2, histo):
        self.r1 = r1
        self.r2 = r2
        self.g1 = g1
        self.g2 = g2
        self.b1 = b1
        self.b2 = b2
        self.histo = histo
        self._avg = None
        self._volumn = None
        self._count = None

    @property
    def volume(self):
        if self._volumn is None:
            sub_r = self.r2 - self.r1
            sub_g = self.g2 - self.g1
            sub_b = self.b2 - self.b1
            self._volumn = (sub_r + 1) * (sub_g + 1) + (sub_b + 1)

        return self._volumn

    @property
    def copy(self):
        copied = deepcopy(self)
        copied._avg = None
        copied._volumn = None
        copied._count = None
        return copied

    @property
    def average(self):
        if self._avg is None:
            total = 0
            mult = 1 << (8 - SIGBITS)
            r_sum = 0
            g_sum = 0
            b_sum = 0
            for i in range(self.r1, self.r2 + 1):
                for j in range(self.g1, self.g2 + 1):
                    for k in range(self.b1, self.b2 + 1):
                        index = get_color_index(i, j, k)
                        hval = self.histo[index]
                        total += hval
                        r_sum += hval * (i + 0.5) * mult
                        g_sum += hval * (j + 0.5) * mult
                        b_sum += hval * (k + 0.5) * mult

            if total:
                r_avg = ~~int(r_sum / total)
                g_avg = ~~int(g_sum / total)
                b_avg = ~~int(b_sum / total)
            else:
                r_avg = ~~int(mult * (self.r1 + self.r2 + 1) / 2)
                g_avg = ~~int(mult * (self.g1 + self.g2 + 1) / 2)
                b_avg = ~~int(mult * (self.b1 + self.b2 + 1) / 2)

            self._avg = r_avg, g_avg, b_avg

        return self._avg

    def contains(self, pixel):
        r_v = pixel[0] >> RSHIFT
        g_v = pixel[1] >> RSHIFT
        b_v = pixel[2] >> RSHIFT

        return all([
            r_v >= self.r1,
            r_v <= self.r2,
            g_v >= self.g1,
            g_v <= self.g2,
            b_v >= self.b1,
            b_v <= self.b2,
        ])

    @property
    def count(self):
        if not self._count:
            n_pix = 0
            for i in range(self.r1, self.r2):
                for j in range(self.g1, self.g2):
                    for k in range(self.b1, self.b2):
                        index_ = get_color_index(i, j, k)
                        try:
                            n_pix += self.histo[index_]
                        except IndexError:
                            pass
            self._count = n_pix
        return self._count



from contextlib import contextmanager
from PIL import Image

from .quantize import mmcq


@contextmanager
def get_palette(filename, color_count=10):
    with Image.open(filename) as image:
        colors = []
        rgb = image.convert('RGB')
        for x in range(0, rgb.width, 5):
            for y in range(0, rgb.height, 5):
                rgb_color = rgb.getpixel((x, y))
                colors.append(rgb_color)

        c_map = mmcq(colors, color_count)
        yield c_map.palette


def get_dominant_color(filename, color_count=10):
    with get_palette(filename, color_count) as palette:
        return palette[0]
